/*
  SSL Proxy Server.
  ----------------------------------------------
  Written by: Eu-Jin Goh (eujin@cs.stanford.edu)
              Stanford University April 2000

 ----------- first the GNU public license -------------
 Copyright (C) 2000  Eu-Jin Goh

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
USA.

--------------------------------------------------

  This program uses the OpenSSL libraries and Dan Boneh's code for the 
  processing of the certificates. Other than that, everything else was 
  coded from scratch.

  The proxy handles SSLV3/TLS servers only. It deals with the first 
  clientHello packet being in SSLV2 format but other than that, no other
  SSLV2 packets are handled. You can typically recognise when the site
  is using SSLV2 when the first byte of the error messages that the
  proxy prints out are -128 or -126. try it on wellsfargo to see.

  This was developed in RedHat linux 6.1 using glibc 2.1.3-8 and openssl
  0.94. Works best in linux using x86 architecture. On some machines, 
  the print outs in hex will appear to have either leading or trailing zeros.

  Please send all bug reports to me. It would be great if you could 
  attach the output when the bug ocurred too. I typically use
  'sslproxy >! out' and read the out file later because there is 
  usually too much output to read at one go.

  NOTE: 
  I have not ported this to big endian architectures yet, i think
  it's not too hard since all that is needed is to change the function
  threeBytesToInt so that it converts the handshake length field to 
  from a three byte number in network byte order to the host byte order.

  If the server returns a cipher suite that is unknown, currently 
  my proxy doesn't handle it too well. it still works but doesn't
  parse the messages properly. 

  Occasionally (very rarely), the multiple application data fragments
  aren't handled properly and invalid type errors keeping appearing
  but the proxy should still work properly after that connection.

  Netscape enterprise servers are rather flakey on the proxy and i added 
  some code so that it works properly. I tested on stronghold servers and
  this proxy works best with those. This is due to the fact that 
  enterprise servers place multiple handshake messages within a single
  TLS record while stronghold sends out handshake messages with their
  own TLS record.

  The application data and certificate messages might be split up over
  several packets and the code that i have currently handles this fine
  but is rather ugly. Given more time, i would rewrite it so that it's 
  cleaner.
  
  THIS SOFTWARE IS PROVIDED BY EU-JIN GOH ``AS IS'' AND ANY EXPRESS OR 
  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <sys/select.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <openssl/x509.h>
#include <openssl/ssl.h>
#include <openssl/bio.h>
#include <openssl/asn1.h>

#ifndef	INADDR_NONE
#define	INADDR_NONE	0xffffffff	/* should be in <netinet/in.h> */
#endif

#define MAX(x,y)  ((x) >= (y) ? (x) : (y))

/* taken from openssl:ssl_locl.h */

#define n2l3(c,l) ((l =(((unsigned long)(c[0]))<<16)| \
		  (((unsigned long)(c[1]))<< 8)| \
		  (((unsigned long)(c[2]))    )),c+=3)

#ifndef true
#define true 1
#endif

#ifndef false
#define false 0
#endif

#define SUCCESS 0
#define FAILURE 1
#define BUFFER_SIZE 2048
#define EXT_BUFFER_SIZE 32768

#define DEFAULT_PORT 8888 /* default if no value is given */

/* for use in doOneTLSConnection */
#define SERVER_RECV_CHANGE_CIPHER 0
#define CLIENT_RECV_CHANGE_CIPHER 1

/* SSL Record Protocol */
/* Content Types */
#define CHANGE_CIPHER_SPEC 20
#define ALERT 21
#define HANDSHAKE 22
#define APPLICATION_DATA 23
#define SSLV2CLIENTHELLO -128

/* Cipher specs */
#define RSA 1
#define DH 2
#define RSA_ENCRYPT_PREMASTER_SECRET_LEN 46

/* Protocol Versions */
#define TLS_MAJOR 3
#define TLS_MINOR 1
#define SSL_MAJOR 3
#define SSL_MINOR 0

/* TLS Alert Protocol msg types */
#define ALERT_LEVEL_WARNING 1
#define ALERT_LEVEL_FATAL 2
#define CLOSE_NOTIFY 0
#define UNEXPECTED_MESSAGE 10
#define BAD_RECORD_MAC 20
#define DECRYPTION_FAILED 21
#define RECORD_OVERFLOW 22
#define DECOMPRESSION_FAILURE 30
#define HANDSHAKE_FAILURE 40
#define BAD_CERTIFICATE 42
#define UNSUPPORTED_CERTIFICATE 43
#define CERTIFICATE_REVOKED 44
#define CERTIFICATE_EXPIRED 45
#define CERTIFICATE_UNKNOWN 46
#define ILLEGAL_PARAMETER 47
#define UNKNOWN_CA 48
#define ACCESS_DENIED 49
#define DECODE_ERROR 50
#define DECRYPT_ERROR 51
#define EXPORT_RESTRICTION 60
#define PROTOCOL_VERSION 70
#define INSUFFICIENT_SECURITY 71
#define INTERNAL_ERROR 80
#define USER_CANCELED 90
#define NO_RENEGOTIATION 100

/* TLS Handshake protocol msg types */
#define HELLO_REQUEST 0
#define CLIENT_HELLO 1
#define SERVER_HELLO 2
#define CERTIFICATE 11
#define SERVER_KEY_EXCHANGE 12
#define CERTIFICATE_REQUEST 13
#define SERVER_HELLO_DONE 14
#define CERTIFICATE_VERIFY 15
#define CLIENT_KEY_EXCHANGE 16
#define FINISHED 20

/* TLS Record Header Definitions */
#define TLS_RECORD_HEADER_SIZE 5
/* used to extract length field from TLS record */
#define TLS_RECORD_LENGTH_OFFSET 3  
/* used to extract encapsulated data from Record */
#define TLS_RECORD_DATA_OFFSET 5

/* Handshake Packet Offsets */
#define HANDSHAKE_DATA_OFFSET 4

/* Handshake Hello Message Offsets from beginning of packet */
#define SESSION_ID_OFFSET 34

/* SSLV2 Client Hello Offsets */
#define CIPHER_SPEC_LEN_OFFSET 3
#define SESSION_ID_LEN_OFFSET 5
#define CHALLENGE_LEN_OFFSET 7
#define CIPHER_SPEC_OFFSET 9

/* Structures used in SSL */

typedef struct protocol_version {
  char major;
  char minor;
}protocol_version;

/*
  dereferencing the short in packet does not seem to work properly
*/
typedef struct tls_packet {
  char content_type;
  struct protocol_version protocol_ver;
  /* unsigned short length;
  char *data;  might want to remove */
}tls_packet;

typedef struct alert {
  char level;
  char description;
}alert;

/* 
   taken from dan boneh's utl_cert.h
   for extracting and printing out the information obtained from the 
   certificate
*/

typedef struct utl_cert_info {

  int keysize;

  char notAfter[64];
  char notBefore[64];

  X509_NAME *subj;
  char subj_DistName[256];

  X509_NAME *issuer;
  char issuer_DistName[256];

} UTL_CERT_INFO;

/* function declarations */
static int initSocketParam(struct sockaddr_in *servAddr, char *port);
static void initPassiveSocket(struct sockaddr_in *servAddr, int listenfd);

static int waitForConnection(int listenfd);
static int waitForClient(struct sockaddr_in *fromAddr, int listenfd);
static int parseCONNECT(char *buffer, struct sockaddr_in *toAddr, 
			 int client_fd);
static void setIP(struct sockaddr_in *servAddr, char *serverHostName);

static void doOneTLSConnection(int client_fd, int server_fd);
static void setfds(fd_set *fds, int server_fd, int client_fd);
static int clientRead(int client_fd, int server_fd, 
		      char *recv_change_cipher, char *keyxchange_alg);
static int serverRead(int client_fd, int server_fd, 
		      char *recv_change_cipher, char *keyxchange_alg);
static char* readAllData(int client_fd, int *buffer_size);
static void handleTLSPacket(char *buffer, int *buffer_size, 
			    char *recv_change_cipher, 
			    char *keyxchange_alg,int client_fd);
static void determineTLSPacketType(tls_packet *packet, char *end, 
				   char *recv_change_cipher,
				   int cur_appl_packet_size);
static int processContentsOfTLSRecord(tls_packet *packet,
				      char *recv_change_cipher,
				      char *keyxchange_alg,
				      int buffer_size,
				      int client_fd);
static void processSSLV2ClientHello(char *buffer);
static void alertMessage(alert *msg);
static int handshakeMessage(char *buffer,char *keyxchange_alg, 
			     int buffer_size,int client_fd);
static void processClientHello(char *buffer);
static int processServerHello(char *buffer, char *keyxchange_alg,
			       int buffer_size,int client_fd);
static char *processHello(char *buffer);
static void processCipherSuite(char byte1, char byte2);
static void keyExchangeAlgorithm(char byte1,char *keyxchange_alg);
static void processSSLV2CipherSuite(char byte1, char byte2, char byte3);
static int processCertificateChain(char *buffer,int record_len,
				    int buffer_size,int client_fd);
static void processCertificate(X509 *x);
static void printCertificateInfo(UTL_CERT_INFO *buf, EVP_PKEY *key);
static void processServerKeyExchange(char *buffer, char *keyxchange_alg);
static char *extractParams(char *params,char *type);
static void processCertificateRequest(char *buffer, char *keyxchange_alg);
static void processClient_key_exchange(char *buffer, char *keyxchange_alg);
static void finishProcessingBuffer
                          (unsigned int data_size, int buffer_size, 
			   tls_packet *packet, int *max_appl_packet_size, 
			   int *cur_appl_packet_size, short *content_length);


char *getMem(int size);
void createSocket(int *sock);
void closeSocket(int sock);
void sendDataTo(int sendfd, char *buffer, int length);
int readDataFrom(int recvfd, char *buffer, int length);
void printCharAsHex(char x);
short twoBytesToInt(char *buffer);
unsigned int threeBytesToInt(char *buffer);

/* main program */
int
main (int argc, char **argv)
{
  struct sockaddr_in servAddr; /* proxy server address */
  int sock;                    /* our socket handler */

  /* sets the port to be the https port is no port is specified. */
  if(argc > 1) 
    sock = initSocketParam(&servAddr,argv[1]);
  else 
    sock = initSocketParam(&servAddr,NULL);

  /* loop till a client connects successfully */
     while(1) waitForConnection(sock); 


  closeSocket(sock);
  return SUCCESS;
}

/*
  Initializes the entries in sockaddr_in struct. Fills in the sin_family,
  sin_port and sin_addr fields if possible.

  Does the socket, bind and listen calls and returns the socket descriptor
  returned by the socket call.

  the socket is set to be reusable.
*/
static int 
initSocketParam(struct sockaddr_in *servAddr, char *port){ 
  short portnum;
  int listenfd;

  /*we get the service port specified by the system if not specified*/
  if(port == NULL)
    portnum = htons(DEFAULT_PORT); 
  else
    portnum = htons(atoi(port));

  createSocket(&listenfd);

  /* now initialize socket parameters */
  memset( (char*) servAddr, 0, sizeof(*servAddr) );
  servAddr->sin_family = AF_INET;         
  servAddr->sin_addr.s_addr = htonl(INADDR_ANY);
  servAddr->sin_port = portnum;

  printf("\nSSLV3/TLS Proxy Server written by Eu-Jin Goh\n"
	 "Stanford University Applied Crypto Group\n");
  printf("\nSSL Proxy Server listening on port number is %d\n\n",ntohs(portnum));

  initPassiveSocket(servAddr,listenfd);

  return listenfd;
}


/*
  Sets up the listening socket for the given port and IP address in servAddr. 
*/
static void
initPassiveSocket(struct sockaddr_in *servAddr, int listenfd)
{
  int sockopt = 1;

  /* sets the socket so that it can be reused. this enables another
     application to make use of the same socket 
  */
  if(setsockopt(listenfd,SOL_SOCKET,SO_REUSEADDR,
		(void *)&sockopt,sizeof(int)) < 0)
    {
      perror("Setting socket options");
      exit(FAILURE);
    }
  
  if(bind(listenfd,(struct sockaddr *)servAddr,sizeof(*servAddr)) < 0)
    {
     perror("bind error");
     exit(FAILURE);
    }
  listen(listenfd,5);
}

/*
  waits till a client connects to the proxy server. It first tries to 
  read from the client socket and parses it. Assumes that the first request
  is a CONNECT command. 
  It then uses the information contained in the request to connect to the 
  server and if successful, replies to the client with a 200 status code.
  Otherwise, it replies with a 400 if the request is not the CONNECT command 
  in the proper format and a 404 if it cannot connect to the dest.

  returns 0 or 1 depending on whether the client connected successfully.
*/
static int
waitForConnection(int listenfd)
{
  int bytes_read=0;       
  int client_fd;           /* socket for the client */
  int server_fd;
  struct sockaddr_in fromAddr;
  struct sockaddr_in toAddr;
  char buffer[BUFFER_SIZE];

  /* waits till a client tries to connect to the proxy server */
  client_fd = waitForClient(&fromAddr, listenfd);

  /* read and print from client socket */
  bytes_read = readDataFrom(client_fd,buffer,BUFFER_SIZE);

  buffer[bytes_read] = '\0';
  printf("\nRead %d bytes from CONNECT request:\n%s",bytes_read,buffer);

  /* obtain dest address and port */
  if(!parseCONNECT(buffer, &toAddr, client_fd)) return 0;   

  /* connect to destination */
  createSocket(&server_fd);  
  if(connect(server_fd,(struct sockaddr *) &toAddr, sizeof(toAddr)) < 0)
    {
      perror("Connect error in waitForConnection");

      sprintf(buffer,"HTTP/1.0 404 Unable to Connect\r\n\r\n");
      printf("Writing to socket: %s",buffer);
      sendDataTo(client_fd, buffer,strlen(buffer));

      return 0;
    }

  /* once connected, reply to client that connection is established */
  sprintf(buffer,"HTTP/1.0 200 Connection Established\r\n\r\n");
  printf("Sending back to client: %s",buffer);
  sendDataTo(client_fd,buffer,strlen(buffer));

  /* handles this TLS Connection */
  doOneTLSConnection(client_fd,server_fd);

  return 1;
}

/*
  accept blocks till a client tries to connect to the proxy server. 
  it then prints out where the connection is originating from 
  (IP and port) and returns the socket descriptor for the connection. 
*/
static int 
waitForClient(struct sockaddr_in *fromAddr, int listenfd)
{
  int connfd;
  int size;
  struct hostent *name;

  size = sizeof(*fromAddr);
  memset((char *)fromAddr,0,size);
  if((connfd = accept(listenfd,(struct sockaddr *) fromAddr,&size)) < 0)
    {
      perror("Accept error in waitForClient");
      exit(FAILURE);
    }

  /* print out where connection originates from */
  name = 
    gethostbyaddr((char *)&(fromAddr->sin_addr),sizeof(fromAddr->sin_addr),
		  AF_INET);
  if(name != NULL)
    {
      printf("--------------------------------------------------------\n");   
      printf("Received connection from %s, port %d\n",
	     name->h_name,ntohs(fromAddr->sin_port));
    }
  else 
    printf("cannot resolve IP of incoming connection\n");
  
  return connfd;
}

/*
  Parses the first message that is sent to the proxy server. It is assumed
  to be in the following format:
  CONNECT IPADDRESS:PORT

  extracts the IPADDRESS and stores it in the sockaddr_in as a binary
  address, together with the port number.

  returns 1 or 0 depending on whether it was successful in parsing. 
*/
int
parseCONNECT(char *buffer, struct sockaddr_in *toAddr, int client_fd)
{
  char *start,*end;
  char *hostname;
  char *port;
  char temp[BUFFER_SIZE];

  start = strchr(buffer,' '); /* location of first space */
  end = strchr(buffer,':');   /* location of ':' */

  if(start == NULL || end == NULL)
    {
      sprintf(temp,"HTTP/1.0 400 Bad Request\r\n\r\n");
      printf("Sending to client: %s",temp);
      sendDataTo(client_fd,temp,strlen(temp));

      return 0;
    }

  /* obtain the hostname */
  hostname = getMem((end - start)/sizeof(char));
  ++start; /* makes it point to the first char */
  strncpy(hostname,start,end-start);
  hostname[end-start] = '\0';

  /* obtain the port number */
  start = strchr(end,' ');
  port = getMem((start - end)/sizeof(char));
  ++end;
  strncpy(port,end,start-end);
  port[start-end] = '\0';

  /* initialize fields in sockaddr_in */
  setIP(toAddr,hostname); /* convert to binary IP */
  toAddr->sin_port = htons( (short) atoi(port));
  toAddr->sin_family = AF_INET;

  printf("Destination hostname is %s, port is %d\n",hostname,ntohs(toAddr->sin_port));
  return 1;
}

/*
  For sin_addr, it first tries to convert it assuming that it is a dotted
  decimal number and if that fails, it uses gethostbyname to convert it. 

  Exits with a failure if unable to resolve the host name.
*/
void 
setIP(struct sockaddr_in *servAddr, char *serverHostName){
  unsigned long	serverInaddr;   /* Binary IP address */
  struct hostent *hostentPtr;   /* used with gethostbyname() */

  /*
    First try to convert the host name as a dotted-decimal number.
    Only if that fails we call gethostbyname().
  */
  if ( (serverInaddr = inet_addr(serverHostName)) != INADDR_NONE )    
    servAddr->sin_addr.s_addr = serverInaddr; /* conversion succeeded */ 
  else 
    {
      if ( (hostentPtr = gethostbyname(serverHostName)) == NULL) 
	{
	  perror("Failed to resolve host name in setIP");
	  exit(FAILURE);
	}
      memcpy((char *) &(servAddr->sin_addr), (char *)hostentPtr->h_addr, 
			hostentPtr->h_length);
    }  
}

/*
  Performs the proxy operations for one TLS connection. Sits in the middle 
  the connection and transfers the data from client to server and vice
  versa. 
  Processes the data that the client and server send to each other to 
  determine at which stage of the TLS protocol the connection is in.
*/

static void 
doOneTLSConnection(int client_fd, int server_fd)
{
  int bytes_read;
  fd_set fds;
  char recv_change_cipher[2];
  char keyxchange_alg = -1; /* to be taken from server hello */


  recv_change_cipher[SERVER_RECV_CHANGE_CIPHER] = 0;
  recv_change_cipher[CLIENT_RECV_CHANGE_CIPHER] = 0;  

  /* sits in this while loop transferring and processing data till one of 
     the sockets close */
  while(true)
    {
      setfds(&fds,server_fd,client_fd);
      switch(select(MAX(client_fd,server_fd)+1,&fds,NULL,NULL,NULL))
	{
	case -1: 
	  perror("Select error"); 
	  exit(FAILURE); 
	default:
	  /* figure out which socket has stuff to read and take the
	     appropriate action */
	  if(FD_ISSET(server_fd,&fds))
	    {
	      printf("\n\nReading from SERVER socket\n");;
	      bytes_read = serverRead(client_fd,server_fd, 
				      recv_change_cipher,
				      &keyxchange_alg);
	    }
	  else if(FD_ISSET(client_fd,&fds))
	    {
	      printf("\n\nReading from CLIENT socket\n");
	      bytes_read = 
		clientRead(client_fd,server_fd,
			   &(recv_change_cipher[CLIENT_RECV_CHANGE_CIPHER]),
			   &keyxchange_alg);
	    }
	}

      if(bytes_read == 0)
	{
	  printf("Zero bytes read, close connections\n\n");
	  closeSocket(server_fd); 
	  closeSocket(client_fd);
	  break;
	}
    }
}

/*
  Sets the fd_set for use in the select function
*/
static void 
setfds(fd_set *fds, int server_fd, int client_fd)
{
  FD_ZERO(fds);
  FD_SET(server_fd,fds);
  FD_SET(client_fd,fds);
}

/*
  Called by doOneTLSConnection to read from the client socket within the 
  select loop. Reads from the client socket and writes the data to the server
  socket.
  
  It reads sockets of up to any size using dynamically allocated memory. 
*/
static int
clientRead(int client_fd, int server_fd, char *recv_change_cipher,
	   char *keyxchange_alg)
{
  int buffer_size = 0;
  char *buffer;

  buffer = readAllData(client_fd,&buffer_size);

  if(buffer_size > 0) 
    {
      /* processes the packet */
      handleTLSPacket(buffer, &buffer_size,recv_change_cipher,
		      keyxchange_alg,client_fd); 
      sendDataTo(server_fd,buffer,buffer_size);
      free(buffer);
    }
  return buffer_size;
}


/*
  Reads from the server socket and writes it to the client one. 
*/
static int 
serverRead(int client_fd, int server_fd, char *recv_change_cipher,
	   char *keyxchange_alg)
{
  return clientRead(server_fd,client_fd, recv_change_cipher,keyxchange_alg);
}


/*
  Called by clientRead. Repeatedly reads from the socket till the number
  of bytes read is less than the max buffer size which implies that all
  the data from the socket has been read. 
*/
static char*
readAllData(int client_fd, int *buffer_size)
{
  int bytes_read;
  char *buffer = getMem(EXT_BUFFER_SIZE+1);
  int max_buffer_size = EXT_BUFFER_SIZE << 1;
  char *larger_buffer = NULL, *temp = NULL;

  /* do first read from the socket */
  bytes_read = readDataFrom(client_fd,buffer,EXT_BUFFER_SIZE);
  *buffer_size = bytes_read;
  
  printf("Read %d bytes...",bytes_read); /* TEST */

  /* in the case where the buffer is not big enough, we enter this
     loop to read the entire contents of the socket */
  while(bytes_read >= EXT_BUFFER_SIZE)
    {
      if(larger_buffer == NULL)
	{
	  larger_buffer = getMem(max_buffer_size);
	  memcpy(larger_buffer,buffer,EXT_BUFFER_SIZE);
	  temp = &(larger_buffer[EXT_BUFFER_SIZE]);
	}
      else 
	{
	  temp = getMem(max_buffer_size);
	  memcpy(temp,larger_buffer,max_buffer_size / 2);
	  larger_buffer = temp;
	  temp = &(larger_buffer[max_buffer_size / 2]);
	}
      bytes_read = readDataFrom(client_fd,temp,EXT_BUFFER_SIZE);
      *buffer_size += bytes_read;
      max_buffer_size = max_buffer_size << 1;      
    }

  /* determine which buffer contains the information required */
  if(larger_buffer == NULL)
    return buffer;
  else 
    return larger_buffer;
}

/*
  handles TLS packets that it reads from the socket. there might be more 
  than one packet in the data that it reads and so it has to handle 
  those cases. Also application data does not come in nice big packets 
  and might be passed to the proxy in fragments and we handle that too.
*/

static void
handleTLSPacket(char *buffer, int *buffer_size, char *recv_change_cipher,
		char *keyxchange_alg,int client_fd)
{
  /* holds state for large application packets */
  static int max_appl_packet_size = 0;
  static int cur_appl_packet_size = 0;

  char end = 0;
  unsigned int data_size = 0;
  short content_length;
  tls_packet *packet= (tls_packet *) buffer;

  while(!end)
    {    
      if(end != -1)
	determineTLSPacketType(packet,&end,recv_change_cipher,
			       cur_appl_packet_size);

      /* if the packet was not one of the standard packets, we do no 
	 further processing */
      if(end > 0) continue;

      /* implies that the packet is a fragment of an application packet,
	 do no further processing */
      if(end == -1)
	{
	  printf("Received APPLICATION DATA packet ...\n"
		 "Fragment of Application data\n");

	  cur_appl_packet_size += *buffer_size;
	  if(cur_appl_packet_size == max_appl_packet_size)
	    {
	      cur_appl_packet_size = 0;
	      max_appl_packet_size = 0;	     
	    }
	  else if(cur_appl_packet_size > max_appl_packet_size)
	    {
	      data_size = *buffer_size - 
		(cur_appl_packet_size - max_appl_packet_size);

	      packet = (tls_packet *) &(buffer[data_size]);

	      cur_appl_packet_size = 0;
	      max_appl_packet_size = 0;	
	      end = 0;
	      printf("\n");
	      continue;
	      }
	  break;
	}

      /* do further processing if it is an alert or handshake message */
      *buffer_size = processContentsOfTLSRecord(packet,recv_change_cipher,
					       keyxchange_alg,
					       *buffer_size,client_fd);

      /* get the length of this TLS packet */
      content_length = twoBytesToInt(((char *)packet) + 
				     TLS_RECORD_LENGTH_OFFSET);

      /* calculate the size of the current packet */
      data_size = data_size + TLS_RECORD_HEADER_SIZE + content_length;

      /* check if there are more packets that were sent with this buffer */
      if(data_size < *buffer_size)
	{
	  printf("\n");
	  packet = (tls_packet *) &(buffer[data_size]); 
	}
      else 
	{
	  finishProcessingBuffer(data_size, *buffer_size, packet,
				 &max_appl_packet_size, 
				 &cur_appl_packet_size, &content_length);
	  break;
	}
      
    }
}

/*
  Determines the kind of packet that has been sent. Handles TLS Record
  Headers. This also handles when application data gets fragmented. the 
  end variable is used to signal when it is a fragmented application data.
*/

static void
determineTLSPacketType(tls_packet *packet, char *end, 
		       char *recv_change_cipher, 
		       int cur_appl_packet_size)
{
  int i;
  char *c = (char *) packet;
  switch((int)packet->content_type)
    {
    case CHANGE_CIPHER_SPEC:
      printf("Received a CHANGE_CIPHER_SPEC packet:\n"
	     "Further packets will be encrypted ... ");

      /* signals that no more processing is to be done */
      *recv_change_cipher = 1;
      break;
    case ALERT:
      printf("Received a ALERT packet ... ");
      break;
    case HANDSHAKE:
      printf("Received a HANDSHAKE packet ...");
      break;
    case APPLICATION_DATA:
      printf("Received APPLICATION DATA packet ...");
      break;
    case SSLV2CLIENTHELLO:
      if(*((char *) packet+2)== CLIENT_HELLO)
	{
	  printf("Received SSLV2 Client Hello ...\n");
	  /* +2 skips over the type and the length field */
	  processSSLV2ClientHello((char *) packet+2); 
	  *end = 1; /* end all further processing */
	  break;
	}
    default:
      /* if cur_appl size is 0 then the previous packet was not 
	 application data because it would have been updated */
      if(cur_appl_packet_size == 0 || (int)packet->content_type == -128)
	{
	  printf("Received Unrecognised TLS packet, type %d\n",
		 packet->content_type);
	  printf("First 30 bytes of data from packet\n");
	  for(i=0;i<30;i++,c++)
	    {
	      printf("%d ",*c);
	    }
	  printf("\n\n");
	  *end = 1; /* trigger to break out of loop */
	}
      else 
	*end = -1; /* mark it as negative to signal that it is a fragment */
      break;
    }
}

/*
  Processes the contents of the data field in the Record. First determines
  the protocol version and then calls the appropriate function if the 
  data in the Record is an Alert or a Handshake message. 

  If the data is encrypted, then no further processing is done. This is 
  signalled by the recv_change_cipher argument

  returns buffer_size.
*/

static int
processContentsOfTLSRecord(tls_packet *packet,char *recv_change_cipher,
			   char *keyxchange_alg, int buffer_size,int client_fd)
{
  /* print out protocol version */
  printf("\nFrom Record Header -- Protocol Version: %d.%d\n",
	 (packet->protocol_ver).major,(packet->protocol_ver).minor);
  
  if((int)packet->content_type == ALERT && !(*recv_change_cipher))
    {
      alertMessage((alert *)((char *)packet + TLS_RECORD_DATA_OFFSET));
    }
  else if((int)packet->content_type == HANDSHAKE 
	  && !(*recv_change_cipher))
    {
      buffer_size = handshakeMessage((char *)packet + TLS_RECORD_DATA_OFFSET,
				     keyxchange_alg, buffer_size,client_fd);
    }
  return buffer_size;
}

/*
  Processes the SSLV2 Client Hello message. This message is only sent by
  clients that support both TLS1.0 and SSLV2. Appendix E RFC 2246.
  Assumes that the buffer points to the message type.
*/

static void 
processSSLV2ClientHello(char *buffer)
{
  short cipher_spec_len = twoBytesToInt(&(buffer[CIPHER_SPEC_LEN_OFFSET]));
  short session_id_len = twoBytesToInt(&(buffer[SESSION_ID_LEN_OFFSET]));
  short challenge_len = twoBytesToInt(&(buffer[CHALLENGE_LEN_OFFSET]));
  char *ptr;
  int session_id_offset = cipher_spec_len + CIPHER_SPEC_OFFSET;
  int i;
  unsigned int byte;

  printf("\nFrom Client Hello -- Protocol Version: %d.%d\n",
	 buffer[1],buffer[2]);

  printf("Session ID Length -- %hd bytes\n",session_id_len);
  printf("Session ID --\n");
  for(i = 0, ptr = buffer + session_id_offset ; 
      i < session_id_len;i++,ptr++)
    {
      byte = 0;
      memcpy(&byte,ptr,1);
      printf("%.2x",byte);
    }

  printf("Cipher Suite Length %hd bytes ... number of cipher suites %d\n",
	 cipher_spec_len,cipher_spec_len/3);
  printf("Cipher Suite List is -- \n");
  for(i = 0, ptr = buffer + CIPHER_SPEC_OFFSET; 
      i < cipher_spec_len; i += 3, ptr += 3)
    processSSLV2CipherSuite(ptr[0],ptr[1],ptr[2]);      
    
  printf("Challenge Length -- %hd bytes\n", challenge_len);
}


/*
  Takes in an alert struct and prints out the level of the alert and also
  the description of the alert.
*/

static void 
alertMessage(alert *msg)
{
  char *str;

  if((int)msg->level == ALERT_LEVEL_WARNING)
    printf("Alert Level: Warning -- ");
  else if((int)msg->level == ALERT_LEVEL_FATAL)
    printf("Alert Level: Fatal -- ");

  switch((int)msg->description)
    {
    case CLOSE_NOTIFY:
      str = "CLOSE_NOTIFY\n"; break;
    case UNEXPECTED_MESSAGE:
      str = "UNEXPECTED_MESSAGE\n"; break;
    case BAD_RECORD_MAC:
      str = "BAD_RECORD_MAC\n"; break;
    case DECRYPTION_FAILED:
      str = "DECRYPTION_FAILED\n"; break;
    case RECORD_OVERFLOW:
      str = "RECORD_OVERFLOW\n"; break;
    case DECOMPRESSION_FAILURE:
      str = "DECOMPRESSION_FAILURE\n"; break;
    case HANDSHAKE_FAILURE:
      str = "HANDSHAKE_FAILURE\n"; break;
    case BAD_CERTIFICATE:
      str = "BAD_CERTIFICATE\n"; break;
    case UNSUPPORTED_CERTIFICATE:
      str = "UNSUPPORTED_CERTIFICATE\n"; break;
    case CERTIFICATE_REVOKED:
      str = "CERTIFICATE_REVOKED\n"; break;
    case CERTIFICATE_EXPIRED:
      str = "CERTIFICATE_EXPIRED\n"; break;
    case CERTIFICATE_UNKNOWN:
      str = "CERTIFICATE_UNKNOWN\n"; break;
    case ILLEGAL_PARAMETER:
      str = "ILLEGAL_PARAMETER\n"; break;
    case UNKNOWN_CA:
      str = "UNKNOWN_CA\n"; break;
    case ACCESS_DENIED:
      str = "ACCESS_DENIED\n"; break;
    case DECODE_ERROR:
      str = "DECODE_ERROR\n"; break;
    case DECRYPT_ERROR:
      str = "DECRYPT_ERROR\n"; break;
    case EXPORT_RESTRICTION:
      str = "EXPORT_RESTRICTION\n"; break;
    case PROTOCOL_VERSION:
      str = "PROTOCOL_VERSION\n"; break;
    case INSUFFICIENT_SECURITY:
      str = "INSUFFICIENT_SECURITY\n"; break;
    case INTERNAL_ERROR:
      str = "INTERNAL_ERROR\n"; break;
    case USER_CANCELED:
      str = "USER_CANCELLED\n"; break;
    case NO_RENEGOTIATION:
      str = "NO_RENEGOTIATION\n"; break;
    default:
      printf("No such alert code %d\n",(int)msg->description);
      return;
    }
  printf("%s\n",str);
}


/*
  Pass in the data segment containing the handshake data extracted from
  the TLS Record. Determines the type and takes the appropriate action.
  Mostly just prints out the Handshake request.
*/

static int
handshakeMessage(char *buffer, char *keyxchange_alg, 
		 int buffer_size,int client_fd)
{
  printf("HandShake Packet Type :- ");

  switch(buffer[0])
    {
    case HELLO_REQUEST:
      printf("Hello Request\n");
      break;
    case CLIENT_HELLO:
      printf("SSLV3/TLS Client Hello\n");
      processClientHello(buffer);
      break;
    case SERVER_HELLO:
      printf("Server Hello\n");
      buffer_size = 
	processServerHello(buffer,keyxchange_alg, buffer_size, client_fd);
      break;
    case CERTIFICATE:
      printf("Certificate\n");
      buffer_size = 
	processCertificateChain(buffer,twoBytesToInt(buffer-2),
				buffer_size,client_fd);
      break;
    case SERVER_KEY_EXCHANGE:
      printf("Server Key Exchange\n");
      processServerKeyExchange(buffer,keyxchange_alg);
      break;
    case CERTIFICATE_REQUEST:
      printf("Certificate request\n");
      processCertificateRequest(buffer,keyxchange_alg);
      break;
    case SERVER_HELLO_DONE:
      printf("Server hello done\n");
      break;
    case CERTIFICATE_VERIFY:
      printf("Certificate verify\n");
      break;
    case CLIENT_KEY_EXCHANGE:
      printf("Client key exchange\n");
      processClient_key_exchange(buffer,keyxchange_alg);
      break;
    case FINISHED:
      printf("Finished Handshake!\n");
      break;
    default:
      printf("Can't recognise Handshake request type %d!\n",buffer[0]);
    }
  return buffer_size;
}

/*
  Given the handshake packet containing the Client Hello message, this
  processes it to obtain the protocol version, session ID and the list of 
  cipher suite that the client can use.
*/

static void 
processClientHello(char *buffer)
{
  char *ptr; /* points at cipher suite */
  short cipher_suite_len;
  int i;

  printf("From Client Hello -- ");
  ptr = processHello(buffer);

  memcpy(&cipher_suite_len,ptr,2);
  cipher_suite_len = ntohs(cipher_suite_len);
  printf("Cipher suite length is %d bytes ... number of cipher suites %d\n",
	 cipher_suite_len,cipher_suite_len/2);

  ptr += 2; /* advance over the length field of cipher suite */
  
  printf("List of cipher suites are -- \n");
  for(i = 0; i < cipher_suite_len; i += 2)    
    processCipherSuite(ptr[i],ptr[i+1]);
    
}

/*
  Given the handshake packet containing the Server Hello message, this
  processes it to obtain the protocol version, session ID and the cipher
  suite that is to be used.

  returns the size of the new buffer_size. might have changed if it read
  more data.
*/
static int
processServerHello(char *buffer,char *keyxchange_alg, int buffer_size,
		   int client_fd)
{
  char *ptr;
  short TLS_record_len = twoBytesToInt(buffer - 2);
  int handshake_record_len = threeBytesToInt(buffer+1);
  int exceed = TLS_record_len  + TLS_RECORD_HEADER_SIZE - buffer_size;
  int temp; 
  char *end; 

  printf("From Server Hello -- ");
  ptr = processHello(buffer); /* points at cipher suite */
  printf("Cipher Suite is -- \n");
  processCipherSuite(ptr[0],ptr[1]);

  keyExchangeAlgorithm(ptr[1],keyxchange_alg);

  /* 
     the following code is to handle netscape enterprise servers. 
     turns out that they bundle many handshake packets into one single
     TLS record. more work for me :-(

     code here is copied from the function handshakeMessage. 
     it's not very pretty or nice.
  */

  /* have to also account for the possiblity that the certificate
     is split over several packets */
  if(exceed > 0)
    {
      end = &(buffer[buffer_size - TLS_RECORD_DATA_OFFSET]);
      buffer_size += exceed; /* going to read this much extra into buffer */
      while(exceed != 0)
	{
	  temp = readDataFrom(client_fd,end,exceed);
	  end += temp;
	  exceed -= temp;
	}
    }
  
  while(TLS_record_len > (handshake_record_len + HANDSHAKE_DATA_OFFSET))
    {
      printf("\nHandShake Packet Type :- ");

      ptr = buffer + HANDSHAKE_DATA_OFFSET + handshake_record_len;

      switch(ptr[0])
	{
	case CERTIFICATE:
	  printf("Certificate\n");
	  /* will have no return value */
	  (void) processCertificateChain(ptr, 0,TLS_RECORD_HEADER_SIZE,
					 client_fd);
	  break;
	case SERVER_KEY_EXCHANGE:
	  printf("Server Key Exchange\n");
	  processServerKeyExchange(ptr,keyxchange_alg);
	  break;
	case CERTIFICATE_REQUEST:
	  printf("Certificate request\n");
	  processCertificateRequest(ptr,keyxchange_alg);
	  break;
	case SERVER_HELLO_DONE:
	  printf("Server hello done\n");
	  handshake_record_len = TLS_record_len; /* to break out */
	  break;
	default:
	  printf("Can't recognise Handshake request type %d!\n",ptr[0]);
	  handshake_record_len = TLS_record_len; /* to break out */	    
	    
	}
      handshake_record_len += 
	threeBytesToInt(ptr+1) + HANDSHAKE_DATA_OFFSET;
    }

  return buffer_size;
}

/*
  Given the handshake packet containing either the ClientHello or ServerHello
  message, it extracts the protocol version, the session ID which is 
  printed byte
  by byte in hex and returns the a pointer to the length field of the cipher 
  suite.
*/
static char *
processHello(char *buffer)
{
  char *temp, *ptr = buffer + HANDSHAKE_DATA_OFFSET;
  int i;
  unsigned int byte;

  /* print out protocol version */
  printf("Protocol Version %d.%d\n",(int)ptr[0],(int)ptr[1]);

  ptr += SESSION_ID_OFFSET; /* skip over random and protocol version field */

  /* sesssion ID in a hello record is only max 32 bytes
     => need 1 byte session ID */
  printf("Length of session ID -- %d bytes\n",(int)(*ptr));

  /* nothing big enough to hold a number of 32 bytes so just iterate 
     over the chars and print them */
  printf("Session ID --\n  0x");
  for(i = 0, temp = ptr + (*ptr) - 1; i < *ptr;i++,temp--)
    {
      byte = 0;
      memcpy(&byte,temp,1);
      printf("%2.2x",byte);
    }
  printf("\n");
  
  /* skip over the number of bytes in the session ID field */
  ptr += ((*ptr) + 1); /* now pointing at cipher suite */

  return ptr;
}

/*
  Examines the bytes given and prints out the cipher suite that is being 
  used. 
*/

static void
processCipherSuite(char byte1,char byte2)
{
  char *name;

  if(byte1 == (char)0xff)
    {
      printf("  Hex Code:");
      printCharAsHex(byte1);
      printCharAsHex(byte2);
      printf("\n  Type: Unknown Cipher Suite\n");
      return;
    }

  /* done because 0x will only be prefixed to a non zero result. */
    printf("  Hex Code:");
    printCharAsHex(byte1);
    printCharAsHex(byte2);
    printf("\n  Type: ");


  /* printf("Int Code: %d Type: ",byte2); */
  switch(byte2)
    {
    case 0x00: 
      name = "No encryption"; 
      break;
    case 0x01: 
      name = "RSA with hash function MD5"; 
      break;
    case 0x02: 
      name = "RSA with SHA"; 
      break;
    case 0x03: 
      name = "RSA EXPORT with 40 bit RC4 and hash function MD5";
      break;
    case 0x04: 
      name = "RSA with 128 bit RC4 and hash function MD5";
      break;
    case 0x05: 
      name = "RSA with 128 bit RC4 and hash function SHA";
      break;
    case 0x06: 
      name = "RSA EXPORT with 40 bit RC2 in CBC mode and hash function MD5"; 
      break;
    case 0x07: 
      name = "RSA with IDEA in CBC mode and hash function SHA";
      break;
    case 0x08:
      name = "RSA EXPORT with 40 bit DES in CBC mode and hash function SHA"; 
      break;
    case 0x09:
      name = "RSA with DES in CBC mode and hash function SHA";
      break;
    case 0x0A:
      name = "RSA with 3DES EDE in CBC mode and hash function SHA";
      break;
    case 0x0B:
      name = "DH DSS EXPORT with 40 bit DES in CBC mode and hash function SHA";
      break;
    case 0x0C:
      name = "DH DSS with DES in CBC mode and hash function SHA";
      break;
    case 0x0D:
      name = "DH DSS with 3DES EDE in CBC mode and hash function SHA";
      break;
    case 0x0E: 
      name = "DH RSA EXPORT with 40 bit DES in CBC mode and hash function SHA";
      break;
    case 0x0F:
      name = "DH RSA with DES in CBC mode and hash function SHA";
      break;
    case 0x10:
      name = "DH RSA with 3DES EDE in CBC mode and hash function SHA";
      break;
    case 0x11: 
      name = "DHE DSS EXPORT with 40 bit DES in CBC mode and hash function SHA";
      break;
    case 0x12:
      name = "DHE DSS with DES in CBC mode and hash function SHA";
      break;
    case 0x13:
      name = "DHE DSS with 3DES EDE in CBC mode and hash function SHA";
      break;
    case 0x14:
      name = "RSA EXPORT with 40 bit DES in CBC mode and hash function SHA";
      break;
    case 0x15:
      name = "DHE RSA with DES in CBC mode and hash function SHA";
      break;
    case 0x16:
      name = "DHE RSA with 3DES EDE in CBC mode and hash function SHA";
      break;
    case 0x17:
      name = "Anonymous DH EXPORT with 40 bit RC4 and hash function MD5";
      break;
    case 0x18:
      name = "Anonymous DH with 128 bit RC4 and hash function MD5";
      break;
    case 0x19:
      name = "Anonymous DH EXPORT with 40 bit DES in CBC mode and hash function SHA"; 
      break;
    case 0x1A:
      name = "Anonymous DH with DES in CBC mode and hash function SHA";
      break;
    case 0x1B:
      name = "Anonymous DH with 3DES EDE in CBC mode and hash function SHA"; 
      break;
      /* Elliptic Curve Cipher Suites */
    case 0x34:
      name = "Elliptic Curve DHE DSS and hash function SHA";
      break;
    case 0x36:
      name = "Elliptic Curve DHE DSS with 128 bit RC4 and hash function SHA";
      break;
    case 0x37:
      name = "Elliptic Curve DHE DSS with DES CBC and hash function SHA";
      break;
    case 0x38:
      name = "Elliptic Curve DHE DSS with 3DES EDE CBC and hash function SHA";
      break;
    case 0x39:
      name = "Elliptic Curve DHE DSS Export with 40 bit DES CBC and hash function SHA";
      break;
    case 0x40:
      name = "Elliptic Curve DHE DSS Export with 40 bit RC4 and hash function SHA";
      break;
    case 0x60:
      name = "RSA Export with 56 bit RC4 and hash function MD5";
      break;
    case 0x61:
      name = "RSA Export with 56 bit RC2 CBC and hash function MD5";
      break;
    case 0x62:
      name = "RSA Export with DES CBC and hash function SHA";
      break;
    case 0x63:
      name = "DHE DSS Export with DES CBC and hash function SHA";
      break;
    case 0x64:
      name = "RSA Export with 56 bit RC4 and hash function SHA";
      break;
    case 0x65:
      name = "DHE DSS Export with 56 bit RC4 and hash function SHA";
      break;
    case 0x66:
      name = "DHE DSS with 128 bit RC4 and hash function SHA";
      break;
    default: 
      printf("Unknown Cipher Suite\n"); 
      return;
    }
  printf("%s\n",name);

}

/*
  Determines if the algorithm is using DH or RSA and returns the value
  through keyxchange_alg.
*/
static void
keyExchangeAlgorithm(char byte1,char *keyxchange_alg)
{
  switch(byte1)
    {
    case 0x01: case 0x02: case 0x03: case 0x04: case 0x05: case 0x06: 
    case 0x07: case 0x08: case 0x09: case 0x0A: case 0x14:
    case 0x60: case 0x61: case 0x62: case 0x64:
      *keyxchange_alg = RSA; break;
    case 0x0B: case 0x0C: case 0x0D: case 0x0E: case 0x0F: case 0x10:
    case 0x11: case 0x12: case 0x13:
    case 0x15: case 0x16: case 0x17: case 0x18: case 0x19: case 0x1A:
    case 0x1B:
    case 0x34: case 0x35: case 0x36: case 0x37: case 0x38: case 0x39:
    case 0x40:
    case 0x63: case 0x65: case 0x66:
      *keyxchange_alg = DH; break; 
    default:
      printf("Error: no such key exchange algorithm\n");
    }
}

/*
  Examines the bytes given and prints out the cipher suite that is being 
  used. This is used when a SSLV2 client hello is being processed.
*/

static void
processSSLV2CipherSuite(char byte1, char byte2, char byte3)
{
  char unknown = 0;

  if(byte1 == 0x00) /* TLS Ciphers used */
    {
      processCipherSuite(byte2,byte3);
      return;
    }

  printf("  Hex Code:");
  printCharAsHex(byte1);
  printCharAsHex(byte2);
  printCharAsHex(byte3);
  printf("\n  Type: ");
  
  switch(byte1)
    {      
    case 0x01:
      if(byte3 == (char) 0x80)
	printf("RSA with 128 bit RC4 and hash function MD5\n");
      else unknown = 1; break;
    case 0x02:
      if(byte3 == (char) 0x80)
	printf("RSA Export with 40 bit RC4 and hash function MD5\n");
      else unknown = 1; break;
    case 0x03:
      if(byte3 == (char) 0x80)
	printf("RSA with 128 bit RC2 CBC and hash function MD5\n");
      else unknown = 1; break;
    case 0x04:
      if(byte3 == (char) 0x80)
	printf("RSA Export with 40 bit RC2 and hash function MD5\n");
      else unknown = 1; break;
    case 0x05:
      if(byte3 == (char) 0x80)
	{
	  if(byte2 == (char) 0x00)
	    printf("RSA with 128 bit IDEA CBC and hash function MD5\n");
	  else
	    printf("RSA with 128 bit IDEA CBC and hash function SHA\n");
	}
      else unknown = 1; break;
    case 0x06:
      if(byte3 == (char) 0x40)
	printf("RSA with 64 bit DES CBC and hash function MD5\n");
      else unknown = 1; break;
    case 0x07:
      if(byte3 == (char) 0xC0)
	{
	  if(byte2 == (char) 0x00)
	    printf("RSA with 192 bit 3DES EDE CBC and hash function MD5\n");
	  else
	    printf("RSA with 192 bit 3DES EDE CBC and hash function SHA\n");
	}
      else unknown = 1; break;
    case 0xff:
      if(byte3 == 0x00)
	printf("RSA with 64 bit DES CFB with hash function MD5\n");
      else
	printf("No Cipher Suite\n");
      break;
    default:
      unknown = 1;
    }
  
  if(unknown)
    printf("Unknown SSLV2 cipher used\n");
}

/*
  processes the certificates that the server might send to the client.
  uses code from the openssl library. 

  returns buffer size in case it reads more from the socket
*/
static int
processCertificateChain(char *buffer,int record_len,int buffer_size,
			int client_fd)
{
  X509 *cert = NULL;
  unsigned long nc,llen,l;
  unsigned char *p,*d,*q;
  int exceed = record_len  + TLS_RECORD_HEADER_SIZE - buffer_size;
  int temp;
  char *end;

  /* have to also account for the possiblity that the certificate
     is split over several packets */
  if(exceed > 0)
    {
      /* means that the record header has a length that exceeds the
	 buffer size by the number of bytes in exceed. 
      */
  
      /* minus the offset because the buffer that we get has that amount
	 added on due to function processContentsOfTLSRecord */
      
      end = &(buffer[buffer_size - TLS_RECORD_DATA_OFFSET]);
      buffer_size += exceed;
      while(exceed != 0)
	{
	  temp = readDataFrom(client_fd,end,exceed);
	  end += temp;
	  exceed -= temp;
	}
    }
  
  d = p = (unsigned char *) (buffer + HANDSHAKE_DATA_OFFSET);
  n2l3(p,llen);

  for (nc = 0; nc < llen; )
    {
      n2l3(p,l);
      if ((l+nc+3) > llen)
	{
	  printf("ERR Certificate length mismatch\n"); 
	  return buffer_size;
	}
      
      q = p;

      cert = d2i_X509(NULL,&q,l); /* grab the current cert */

      if (cert == NULL)
	{
	  printf("ERR Bad Certificate\n"); 
	  return buffer_size;
	}
      if (q != (p+l))
	{
	  printf("ERR in Certificate Decode\n"); 
	  return buffer_size;
	}
      
      processCertificate(cert);

      cert = NULL;
      nc += l+3;
      p = q;
    }
  return buffer_size;
}

/*
  takes the X509 certificate and parses out the information into a 
  UTL_CERT_INFO struct. 
  Taken from Dan Boneh's utl_cert.c
*/

static void
processCertificate(X509 *x)
{
  UTL_CERT_INFO buf;
  BIO *mem = NULL;
  EVP_PKEY *key;

  if ((mem=BIO_new(BIO_s_mem())) == NULL)
    {
      printf("ERR Unable to create new BIO\n");
      return;
    }

  /* Extract and process validity periods */
  ASN1_UTCTIME_print(mem, X509_get_notAfter(x));
  BIO_gets(mem, buf.notAfter, sizeof(buf.notAfter));
  ASN1_UTCTIME_print(mem, X509_get_notBefore(x));
  BIO_gets(mem, buf.notBefore, sizeof(buf.notBefore));

  /* Extract and process subject name */
  buf.subj = X509_get_subject_name(x);
  X509_NAME_oneline(buf.subj,
		    buf.subj_DistName, sizeof(buf.subj_DistName) );

  /* Extract and process issuer name */
  buf.issuer = X509_get_issuer_name(x);
  X509_NAME_oneline(buf.issuer,
		    buf.issuer_DistName, sizeof(buf.issuer_DistName) );

  /* get the key size */
  key = X509_get_pubkey(x);

  printCertificateInfo(&buf,key);

  BIO_free(mem);
  EVP_PKEY_free(key);
  
}

/*
  takes the info from utl_cert_info and prints out the info. 
  also prints out the key size and type of key exchange mechanism.
  key handling taken from Dan Boneh's utl_cert.c
*/

static void
printCertificateInfo(UTL_CERT_INFO *buf, EVP_PKEY *key)
{
  printf("  CERTIFICATE INFORMATION :- \n");
  printf("  Validity -- Not After  %s\n",buf->notAfter);
  printf("              Not Before %s\n",buf->notBefore);
  printf("  Subject Distinguished Name -- \n    %s\n",buf->subj_DistName);
  printf("  Issuer Distinguished Name  -- \n    %s\n",buf->issuer_DistName);

  if (key == NULL) return;

  switch (key->type) 
    {
    case EVP_PKEY_RSA:
      buf->keysize = RSA_size(key->pkey.rsa)*8;
      printf("  RSA Public key size %d bits\n\n",buf->keysize);
      break;      
    case EVP_PKEY_DSA:
      buf->keysize = DSA_size(key->pkey.dsa)*8;
      printf("  DSS Public key size %d bits\n\n",buf->keysize);
      break;
    default:
      printf("  Unknown key type\n\n");
    }
}

/*
  processes the server key exchange to obtain the parameters. 
  depending on whether the algorithm is RSA or DSS, different
  params will be printed out.
*/
static void 
processServerKeyExchange(char *buffer, char *keyxchange_alg)
{
  char *data = buffer + HANDSHAKE_DATA_OFFSET;

  switch(*keyxchange_alg)
    {
    case RSA:
      data = extractParams(data,"RSA Modulus");
      extractParams(data,"RSA Exponent");
      break;
    case DH:
      data = extractParams(data,"DH Prime Modulus p");
      data = extractParams(data,"DH Generator g");
      data = extractParams(data,"DH public value (g^X mod p)");
      break;
    }
}

/*
  prints out the data in hex of the parameters. assumes that the length
  of the data is given by the first 2 bytes.
*/

static char *
extractParams(char *params,char *type)
{
  unsigned short param_len = twoBytesToInt(params);
  char *start = params;
  int i;  
  unsigned int byte;

  printf("Length of %s -- %d\n",type,param_len);
  printf("%s --\n  0x",type);
  for(i = 0, params += 1 + param_len; i < param_len; i++,params--)
    {
      byte = 0; memcpy(&byte,params,1);
      printf("%2.2x",byte);		
    }
  printf("\n"); 
  return (start+2+param_len);
}

/*
  prints out the types of certificates requested. At present just 
  dumps out the distinguished names of the CA as a string.
  -------------------
  might have to figure out the distinguished names format and print
  nicely
*/

static void 
processCertificateRequest(char *buffer, char *keyxchange_alg)
{
  char *data = buffer + HANDSHAKE_DATA_OFFSET;
  char *temp;
  unsigned int certTypeLen = 0;
  int i;  
  unsigned int len = threeBytesToInt(buffer+1);

  memcpy(&certTypeLen,data,1);
  printf("Types of certificates requested --\n");

  for(i = 0, data++; i < certTypeLen; i++,data++)
    {
      switch(*data)
	{
	case 1:
	  printf("\tRSA certificate\n"); break;	  
	case 2:
	  printf("\tDSS certificate\n"); break;
	case 3:
	  printf("\tRSA certificate with fixed DH parameters\n"); break;
	case 4:
	  printf("\tDSS certificate with fixed DH parameters\n"); break;
	default:
	  printf("\tUnknown certificate requested\n"); break;
	}
    }
  
  /* print out the distinguished names of the CA */
  len -= certTypeLen;
  temp = getMem(len+1);
  memcpy(temp,data,len);
  temp[len] = '\0';
  printf("Distinguished Names of CA's -- %s",temp);

  free(temp);
  
}

/*
  Using the keyxchange_alg arg passed in, will process and print 
  out the appropriate information. For RSA, prints out the RSA encrypted
  premaster secret and for DH, prints out the public value.
  ------------
  DH part is untested
*/
static void 
processClient_key_exchange(char *buffer, char *keyxchange_alg)
{
  char *ptr = buffer + HANDSHAKE_DATA_OFFSET;
  int i; 
  unsigned int len = threeBytesToInt(buffer+1);
  unsigned int byte = 0;

  switch(*keyxchange_alg)
    {
    case RSA:
      printf("Length of RSA Encrypted PreMaster Secret -- %d bytes\n",len);
      
      printf("RSA Encrypted PreMaster Secret --\n  0x");
      for(i = 0, ptr += 1 + len; i < len; i++,ptr--)
	{
	  byte = 0; memcpy(&byte,ptr,1);
	  printf("%2.2x",byte);		
	}
      printf("\n");
      break;
    case DH:
        if(!len)	
	  extractParams(ptr,"DH public value");	
	else
	  printf("DH public key is implicit in client certificate\n");
	break;
    default:
      printf("Unable to process Client Key Exchange %d\n",*keyxchange_alg);
    }
}


/*
  Called by handleTLSPacket when the data has been completely processed.
  Determines if the processing was successful and whether it was a 
  fragment of application data.
*/

static void
finishProcessingBuffer(unsigned int data_size, int buffer_size, 
		       tls_packet *packet, int *max_appl_packet_size, 
		       int *cur_appl_packet_size, short *content_length)
{
  if(data_size == buffer_size)
    printf("Finished processing all data.\n"); /* TEST */
  else if(packet->content_type == APPLICATION_DATA)
    {
      printf("Fragment of Application Data.\n");
      *max_appl_packet_size = 
	*content_length + TLS_RECORD_DATA_OFFSET;
      *cur_appl_packet_size = buffer_size;
    }
  else
    printf("Error: Incorrect data length given in Record\n");
}


/*
  Just calls a malloc except that it checks that the memory has actually
  been allocated. Exits from program if failure to allocate memory.
  Used to allocate memory for char *.
*/
char *
getMem(int size){

  char *temp = (char *) malloc(size);

  if (!temp)
    {
      perror("Malloc failed to allocate memory!");
      exit(FAILURE);
    } 
  else 
    return temp;
}

/*
  Creates a stream socket. the argument purpose is for printing out the
  error message.
*/
void 
createSocket(int *sock)
{
  if ( ( *sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) 
    {
      perror("Failed to open stream socket");
      exit(FAILURE);
    }
}

/*
  A wrapper for close that does some error checking.
*/
void 
closeSocket(int sock)
{
  if(close(sock) < 0 )
    {      
      perror("Close socket error");
      exit(FAILURE);
    }
}

/*
  reads the data from the socket to the buffer
*/
int
readDataFrom(int recvfd, char *buffer, int length)
{
  int bytes_read;

  if((bytes_read = read(recvfd,buffer,length)) < 0)
    {
      perror("Error reading in readDataFrom");
      exit(FAILURE);
    }
  return bytes_read;
}

/*
  writes the data in the buffer to the socket.
*/
void 
sendDataTo(int sendfd, char *buffer, int length)
{ 
  if(write(sendfd,buffer,length) < 0)
    {
      perror("Write error in sendDataTo");
      exit(FAILURE);
    }
}

/*
  Prints a char as hex. Converts the char to a unsigned int to avoid 
  sign extensions. This is used because printf does not support 
  printing 0x if the value is 0.
*/
void
printCharAsHex(char x)
{
  unsigned int byte=0;

  memcpy(&byte,&x,1);

  if(byte) printf(" %#.2x",byte);
  else printf(" 0x00");
}

/*
  used for converting the short in 2 byte length fields to an int. 
  returns the value as a short in the host byte order.
*/

short
twoBytesToInt(char *buffer)
{
  int i;  
  char *temp;
  short len=0;

  /* get the length of the message */
  temp = (char *) &len;
  for(i = 0;i < 2; i++) /* 2 is the size of the length field */
    {
      memcpy(&(temp[i]),buffer+i,1);
    }

  return ntohs(len);
}

/*
  used for converting the 3 byte field in the handshake packet that
  represents the length of the packet to an int
*/
unsigned int
threeBytesToInt(char *buffer)
{
  int i;  
  char *temp;
  unsigned int len=0;

  /* get the length of the message */
  temp = (char *) &len;
  for(i = 0;i < 3; i++) /* 3 is the size of the length field */
    {
      memcpy(&(temp[i]),buffer+2-i,1);
    }

  return len;
}
